#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import copy

from pydbal.exception import DBALBuilderError


class SQLBuilder:
    SELECT = 0
    DELETE = 1
    UPDATE = 2
    INSERT = 3

    STATE_DIRTY = 0
    STATE_CLEAN = 1

    def __init__(self, connection):
        self._connection = connection
        self._sql_parts = {
            "select":   [],
            "from":     [],
            "join":     {},
            "set":      [],
            "where":    None,
            "group_by": [],
            "having":   None,
            "order_by": [],
            "values":   {}
        }
        self._sql = None
        self._params = {}
        self._type = SQLBuilder.SELECT
        self._state = SQLBuilder.STATE_CLEAN
        self._first_result = None
        self._max_results = None
        self._param_counter = 0

    def __str__(self):
        return self.get_sql()

    def copy(self):
        return copy.copy(self)

    def expr(self):
        return self._connection.get_expression_builder()

    def get_type(self):
        return self._type

    def get_connection(self):
        return self._connection

    def set_parameter(self, key, value):
        if isinstance(key, str):
            key = key.lstrip(":")
        elif not isinstance(key, int):
            raise ValueError("Argument 'key' must be int or string.")
        self._params[key] = value
        return self

    def set_parameters(self, params):
        self._params.clear()
        if isinstance(params, dict):
            for key, param in params.iteritems():
                self.set_parameter(key, param)
        elif isinstance(params, (list, tuple)):
            for i, param in enumerate(params):
                self.set_parameter(i, param)
        else:
            raise ValueError("Argument 'params' must be dict, list or tuple.")
        return self

    def get_parameter(self, key):
        return self._params[key]

    def get_parameters(self):
        return self._params

    def set_first_result(self, first_result):
        self._state = SQLBuilder.STATE_DIRTY
        self._first_result = first_result
        return self

    def get_first_result(self):
        return self._first_result

    def set_max_results(self, max_results):
        self._state = SQLBuilder.STATE_DIRTY
        self._max_results = max_results
        return self

    def get_max_results(self):
        return self._max_results

    def _add(self, sql_part_name, sql_part, append=False):
        if not sql_part:
            return self

        self._state = SQLBuilder.STATE_DIRTY

        if not append:
            self.reset_sql_part(sql_part_name)

        if sql_part_name in ("select", "group_by"):
            self._sql_parts[sql_part_name].extend(sql_part)
        elif sql_part_name in ("from", "set", "order_by"):
            self._sql_parts[sql_part_name].append(sql_part)
        elif sql_part_name == "join":
            if sql_part[0] in self._sql_parts["join"]:
                self._sql_parts["join"][sql_part[0]].append(sql_part[1:])
            else:
                self._sql_parts["join"][sql_part[0]] = [sql_part[1:]]
        else:
            self._sql_parts[sql_part_name] = sql_part
        return self

    def select(self, select, *args):
        self._type = SQLBuilder.SELECT
        return self._add("select", (select, ) + args)

    def add_select(self, select, *args):
        self._type = SQLBuilder.SELECT
        return self._add("select", (select, ) + args, True)

    def from_(self, table, alias=None):
        return self._add("from", (table, alias), True)

    def insert(self, table):
        self._type = SQLBuilder.INSERT
        return self._add("from", (table, ))

    def update(self, table, alias=None):
        self._type = SQLBuilder.UPDATE
        return self._add("from", (table, alias))

    def delete(self, table, alias=None):
        self._type = SQLBuilder.DELETE
        return self._add("from", (table, alias))

    def inner_join(self, from_alias, join, alias, *condition):
        condition = CompositeExpression(CompositeExpression.TYPE_AND, *condition)
        return self._add("join", (from_alias, "inner", join, alias, condition), True)

    def left_join(self, from_alias, join, alias, *condition):
        condition = CompositeExpression(CompositeExpression.TYPE_AND, *condition)
        return self._add("join", (from_alias, "left", join, alias, condition), True)

    def right_join(self, from_alias, join, alias, *condition):
        condition = CompositeExpression(CompositeExpression.TYPE_AND, *condition)
        return self._add("join", (from_alias, "right", join, alias, condition), True)

    join = inner_join

    def set(self, key, value):
        return self._add("set", key + " = " + value, True)

    def where(self, where, *args):
        return self._add("where", CompositeExpression(CompositeExpression.TYPE_AND, *(where, ) + args))

    def and_where(self, where, *args):
        where = (where, ) + args
        expr = self._sql_parts["where"]
        if isinstance(expr, CompositeExpression):
            where = (str(expr), ) + where
        return self._add("where", CompositeExpression(CompositeExpression.TYPE_AND, *where))

    def or_where(self, where, *args):
        where = (where, ) + args
        expr = self._sql_parts["where"]
        if isinstance(expr, CompositeExpression):
            where = (str(expr), ) + where
        return self._add("where", CompositeExpression(CompositeExpression.TYPE_OR, *where))

    def group_by(self, group_by, *args):
        return self._add("group_by", (group_by, ) + args)

    def add_group_by(self, group_by, *args):
        return self._add("group_by", (group_by, ) + args, True)

    def set_value(self, column, value):
        self._sql_parts["values"][column] = value
        return self

    def values(self, values):
        if isinstance(values, dict):
            return self._add("values", values)
        return self

    def having(self, having, *args):
        return self._add("having", CompositeExpression(CompositeExpression.TYPE_AND, *(having, ) + args))

    def and_having(self, having, *args):
        having = (having, ) + args
        expr = self._sql_parts["having"]
        if isinstance(expr, CompositeExpression):
            having = (str(expr), ) + having
        return self._add("having", CompositeExpression(CompositeExpression.TYPE_AND, *having))

    def or_having(self, having, *args):
        having = (having, ) + args
        expr = self._sql_parts["having"]
        if isinstance(expr, CompositeExpression):
            having = (str(expr), ) + having
        return self._add("having", CompositeExpression(CompositeExpression.TYPE_OR, *having))

    def order_by(self, sort, order="ASC"):
        return self._add("order_by", sort + " " + order.upper())

    def add_order_by(self, sort, order="ASC"):
        return self._add("order_by", sort + " " + order.upper(), True)

    def reset_sql_parts(self, sql_part_names=None):
        if sql_part_names is None:
            sql_part_names = self._sql_parts.keys()
        for sql_part_name in sql_part_names:
            self.reset_sql_part(sql_part_name)
        return self

    def reset_sql_part(self, sql_part_name):
        if isinstance(self._sql_parts[sql_part_name], list):
            del self._sql_parts[sql_part_name][:]
        elif isinstance(self._sql_parts[sql_part_name], dict):
            self._sql_parts[sql_part_name].clear()
        else:
            self._sql_parts[sql_part_name] = None
        self._state = SQLBuilder.STATE_DIRTY
        return self

    def create_named_parameter(self, value, placeholder=None):
        if placeholder is None:
            placeholder = ":pyValue" + str(self._param_counter)
            self._param_counter += 1
        self.set_parameter(placeholder[1:], value)
        return placeholder

    def create_positional_parameter(self, value):
        self.set_parameter(self._param_counter, value)
        self._param_counter += 1
        return "?"

    def get_sql(self):
        if self._sql is not None and self._state == SQLBuilder.STATE_CLEAN:
            return self._sql

        if self._type == SQLBuilder.INSERT:
            self._sql = self._get_sql_for_insert()
        elif self._type == SQLBuilder.DELETE:
            self._sql = self._get_sql_for_delete()
        elif self._type == SQLBuilder.UPDATE:
            self._sql = self._get_sql_for_update()
        else:
            self._sql = self._get_sql_for_select()

        self._state = SQLBuilder.STATE_CLEAN
        return self._sql

    def _get_sql_for_select(self):
        sql = "SELECT " + ", ".join(self._sql_parts["select"]) + " FROM "
        sql += ", ".join(self._get_from_clauses().itervalues())
        if self._sql_parts["where"] is not None:
            sql += " WHERE " + str(self._sql_parts["where"])
        if self._sql_parts["group_by"]:
            sql += " GROUP BY " + ", ".join(self._sql_parts["group_by"])
        if self._sql_parts["having"] is not None:
            sql += " HAVING " + str(self._sql_parts["having"])
        if self._sql_parts["order_by"]:
            sql += " ORDER BY " + ", ".join(self._sql_parts["order_by"])

        if self._max_results is not None or self._first_result is not None:
            return self._connection.get_platform().modify_limit_sql(sql, self._max_results, self._first_result)
        return sql

    def _get_from_clauses(self):
        from_clauses = {}
        known_aliases = set()
        for from_ in self._sql_parts["from"]:
            if from_[1] is None:
                table_sql = from_[0]
                table_reference = from_[0]
            else:
                table_sql = from_[0] + " " + from_[1]
                table_reference = from_[1]

            known_aliases.add(table_reference)
            from_clauses[table_reference] = table_sql + self._get_sql_for_joins(table_reference, known_aliases)

        for from_alias in self._sql_parts["join"].iterkeys():
            if from_alias not in known_aliases:
                raise DBALBuilderError.unknown_alias(from_alias, known_aliases)
        return from_clauses

    def _get_sql_for_joins(self, from_alias, known_aliases):
        sql = ""
        if from_alias in self._sql_parts["join"]:
            for join in self._sql_parts["join"][from_alias]:
                if join[2] in known_aliases:
                    raise DBALBuilderError.non_unique_alias(join[2], known_aliases)
                sql += " %s JOIN %s %s ON %s" % (
                    join[0].upper(),
                    join[1],
                    join[2],
                    str(join[3])
                )
                known_aliases.add(join[2])

            for join in self._sql_parts["join"][from_alias]:
                sql += self._get_sql_for_joins(join[2], known_aliases)
        return sql

    def _get_sql_for_insert(self):
        return "INSERT INTO %s (%s) VALUES(%s)" % (
            self._sql_parts["from"][0][0],
            ", ".join(self._sql_parts["values"].iterkeys()),
            ", ".join(self._sql_parts["values"].itervalues())
        )

    def _get_sql_for_update(self):
        sql = "UPDATE " + self._sql_parts["from"][0][0]
        if self._sql_parts["from"][0][1] is not None:
            sql += " " + self._sql_parts["from"][0][1]
        if self._sql_parts["set"]:
            sql += " SET " + ", ".join(self._sql_parts["set"])
        if self._sql_parts["where"] is not None:
            sql += " WHERE " + str(self._sql_parts["where"])
        return sql

    def _get_sql_for_delete(self):
        sql = "DELETE FROM " + self._sql_parts["from"][0][0]
        if self._sql_parts["from"][0][1] is not None:
            sql += " " + self._sql_parts["from"][0][1]
        if self._sql_parts["where"] is not None:
            sql += " WHERE " + str(self._sql_parts["where"])
        return sql

    def _prepare_params(self):
        args, kwargs = [], {}
        for key, value in self._params.iteritems():
            if isinstance(key, int):
                args.append(value)
            else:
                kwargs[key] = value
        return args, kwargs

    def execute(self):
        args, kwargs = self._prepare_params()
        if self._type == SQLBuilder.SELECT:
            return self._connection.query(self.get_sql(), *args, **kwargs)
        result = self._connection.execute(self.get_sql(), *args, **kwargs)
        if self._type == SQLBuilder.INSERT:
            return self._connection.last_insert_id()
        return result


class ExpressionBuilder:
    EQ = "="
    NEQ = "<>"
    LT = "<"
    LTE = "<="
    GT = ">"
    GTE = ">="
    IS_NULL = "IS NULL"
    IS_NOT_NULL = "IS NOT NULL"
    LIKE = "LIKE"
    NOT_LIKE = "NOT LIKE"
    IN = "IN"
    NOT_IN = "NOT IN"

    def __init__(self, connection):
        self._connection = connection

    @staticmethod
    def and_x(*x):
        return CompositeExpression(CompositeExpression.TYPE_AND, *x)

    @staticmethod
    def or_x(*x):
        return CompositeExpression(CompositeExpression.TYPE_OR, *x)

    @staticmethod
    def comparison(x, operator, y):
        return x + " " + operator + " " + y

    @staticmethod
    def eq(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.EQ, y)

    @staticmethod
    def neq(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.NEQ, y)

    @staticmethod
    def lt(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.LT, y)

    @staticmethod
    def lte(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.LTE, y)

    @staticmethod
    def gt(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.GT, y)

    @staticmethod
    def gte(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.GTE, y)

    @staticmethod
    def is_null(x):
        return x + " " + ExpressionBuilder.IS_NULL

    @staticmethod
    def is_not_null(x):
        return x + " " + ExpressionBuilder.IS_NOT_NULL

    @staticmethod
    def like(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.LIKE, y)

    @staticmethod
    def not_like(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.NOT_LIKE, y)

    @staticmethod
    def in_(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.IN, "(" + ", ".join(y) + ")")

    @staticmethod
    def not_in(x, y):
        return ExpressionBuilder.comparison(x, ExpressionBuilder.NOT_IN, "(" + ", ".join(y) + ")")

    def literal(self, value):
        return self._connection.get_driver().escape_string(value)


class CompositeExpression:
    TYPE_AND = "AND"
    TYPE_OR = "OR"

    def __init__(self, type_, *parts):
        self._type = type_
        self._parts = []
        self.add_multiple(parts)

    def __len__(self):
        return len(self._parts)

    def __str__(self):
        if len(self._parts) == 1:
            return CompositeExpression._str(self._parts[0])
        parts = map(CompositeExpression._str, self._parts)
        return "(" + (") " + self._type + " (").join(parts) + ")"

    def __iter__(self):
        return iter(self._parts)

    @staticmethod
    def _str(part):
        if isinstance(part, CompositeExpression):
            return str(part)
        return part

    def copy(self):
        return copy.copy(self)

    def get_type(self):
        return self._type

    def add_multiple(self, parts=None):
        if parts is None:
            parts = []
        for part in parts:
            self.add(part)
        return self

    def add(self, part):
        if (isinstance(part, CompositeExpression) and len(part) > 0) or part:
            self._parts.append(part)
        return self



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

from abc import ABCMeta, abstractmethod
from binascii import crc32

from pydbal.types import BaseType
from pydbal.cache import cached


class BaseAsset:
    __metaclass__ = ABCMeta

    _name = None
    _namespace = None
    _quoted = False

    def __repr__(self):
        return "<%s.%s> %s" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.__str__()
        )

    @abstractmethod
    def __str__(self):
        pass

    def _set_name(self, name):
        if BaseAsset._is_identifier_quoted(name):
            self._quoted = True
            name = BaseAsset._trim_quotes(name)
        if "." in name:
            self._namespace, name = name.split(".", 1)
        self._name = name

    def get_name(self):
        if self._namespace:
            return self._namespace + "." + self._name
        return self._name

    def get_namespace(self):
        return self._namespace

    @staticmethod
    def _is_identifier_quoted(identifier):
        return identifier[:1] in ("`", '"', "[")

    @staticmethod
    def _trim_quotes(identifier):
        return identifier.encode("utf-8").translate(None, '`"[]')

    def _get_quoted_name(self, platform):
        keywords = platform.get_keywords()

        def quote(identifier):
            if self._quoted or identifier in keywords:
                return platform.quote_single_identifier(identifier)
            return identifier
        return ".".join(map(quote, self.get_name().split(".")))

    @staticmethod
    def _generate_identifier_name(column_names, prefix="", max_size=30):
        def encode(column_name):
            return "%X" % (crc32(column_name) & 0xffffffff)
        return (prefix.upper() + "_" + "".join(map(encode, column_names)))[:max_size]


class View(BaseAsset):
    def __init__(self, name, sql):
        self._set_name(name)
        self._sql = sql

    def __str__(self):
        return "'%s'" % self.get_name()

    def get_sql(self):
        return self._sql


class Table(BaseAsset):
    def __init__(self, name, columns=None, indexes=None, foreign_keys=None, **options):
        self._set_name(name)
        self._columns = set(columns)
        self._indexes = set(indexes)
        self._foreign_keys = set(foreign_keys)
        self._options = options

    def __str__(self):
        return "'%s'" % self.get_name()

    def get_columns(self):
        return tuple(self._columns)

    def get_indexes(self):
        return tuple(self._indexes)

    def get_foreign_keys(self):
        return tuple(self._foreign_keys)

    def get_options(self):
        return self._options.copy()


class Column(BaseAsset):
    def __init__(self, name, type_, length=None, precision=10, scale=0, unsigned=False, fixed=False, notnull=True,
                 default=None, autoincrement=False, column_definition=None, comment=None, platform_options=None,
                 **custom_schema_options):
        assert issubclass(type_, BaseType)

        self._set_name(name)
        self._type = type_

        if length is not None:
            length = int(length)
        self._length = length

        self._precision = int(precision)
        self._scale = int(scale)
        self._unsigned = bool(unsigned)
        self._fixed = bool(fixed)
        self._notnull = bool(notnull)
        self._default = default
        self._autoincrement = bool(autoincrement)

        self._column_definition = str(column_definition)
        self._comment = str(comment)

        self._platform_options = dict(platform_options) if platform_options else {}
        self._custom_schema_options = custom_schema_options

    def __str__(self):
        return "'%s' [%s]" % (self.get_name(), self._type.get_name())

    def get_type(self):
        return self._type

    def get_length(self):
        return self._length

    def get_precision(self):
        return self._precision

    def get_scale(self):
        return self._scale

    def is_unsigned(self):
        return self._unsigned

    def is_fixed(self):
        return self._fixed

    def is_notnull(self):
        return self._notnull

    def get_default(self):
        return self._default

    def is_autoincrement(self):
        return self._autoincrement

    def get_platform_options(self):
        return self._platform_options.copy()

    def get_column_definition(self):
        return self._column_definition

    def get_comment(self):
        return self._comment

    def get_custom_schema_options(self):
        return self._custom_schema_options.copy()


class Index(BaseAsset):
    def __init__(self, name, columns, unique=False, primary=False, flags=None, **options):
        self._set_name(name)
        self._columns = set(columns)

        self._unique = unique or primary
        self._primary = primary

        self._flags = set(flags or ())
        self._options = options

    def __str__(self):
        return "'%s' %s" % (self.get_name(), list(self._columns))

    def get_columns(self):
        return tuple(self._columns)

    def is_unique(self):
        return self._unique

    def is_primary(self):
        return self._primary

    def get_flags(self):
        return tuple(self._flags)

    def get_options(self):
        return self._options.copy()


class ForeignKey(BaseAsset):
    def __init__(self, name, local_columns, foreign_table, foreign_columns, **options):
        self._set_name(name)
        self._local_columns = set(local_columns)
        self._foreign_table = foreign_table
        self._foreign_columns = set(foreign_columns)
        self._options = options

    def __str__(self):
        return "`%s` %s -> '%s'.%s" % (
            self.get_name(),
            list(self._local_columns),
            self._foreign_table,
            list(self._foreign_columns)
        )

    def get_local_columns(self):
        return tuple(self._local_columns)

    def get_foreign_table(self):
        return self._foreign_table

    def get_foreign_columns(self):
        return tuple(self._foreign_columns)

    def get_options(self):
        return self._options.copy()


class SchemaManager:
    def __init__(self, connection):
        self._connection = connection
        self._platform = connection.get_platform()

    def __contains__(self, item):
        if isinstance(item, (list, tuple)):
            return all(x in self for x in item)
        try:
            name = item.get_name().lower()
        except AttributeError:
            name = item.lower()
        if isinstance(item, (Table, basestring)):
            return name in (x.lower() for x in self.get_table_names())
        elif isinstance(item, View):
            return name in (x.lower() for x in self.get_view_names())
        return False

    @cached
    def get_database_names(self):
        self._connection.ensure_connected()
        return list(self._platform.get_databases())

    @cached
    def get_views(self, database=None):
        self._connection.ensure_connected()
        return map(lambda x: View(*x), self._platform.get_views(database))

    def get_view_names(self, database=None, **kwargs):
        return map(View.get_name, self.get_views(database, **kwargs))

    @cached
    def get_table(self, table, database=None):
        columns = self.get_table_columns(table, database)
        indexes = self.get_table_indexes(table, database)

        foreign_keys = None
        if self._platform.is_foreign_keys_supported():
            foreign_keys = self.get_table_foreign_keys(table, database)

        return Table(table, columns, indexes, foreign_keys)

    @cached
    def get_table_names(self, database=None):
        self._connection.ensure_connected()
        return list(self._platform.get_tables(database))

    @cached
    def get_table_columns(self, table, database=None):
        self._connection.ensure_connected()

        def create_column(definition):
            return Column(definition[0], BaseType.get_type(definition[1]), **definition[2])

        return map(create_column, self._platform.get_table_columns(table, database))

    def get_table_column_names(self, table, database=None, **kwargs):
        return map(Column.get_name, self.get_table_columns(table, database, **kwargs))

    @cached
    def get_table_indexes(self, table, database=None):
        self._connection.ensure_connected()

        def create_index(definition):
            return Index(definition[0], definition[1], **definition[2])

        return map(create_index, self._platform.get_table_indexes(table, database))

    def get_table_index_names(self, table, database=None, **kwargs):
        return map(Index.get_name, self.get_table_indexes(table, database, **kwargs))

    @cached
    def get_table_foreign_keys(self, table, database=None):
        self._connection.ensure_connected()

        def create_foreign_key(definition):
            return ForeignKey(definition[0], definition[1], definition[2], definition[3], **definition[4])

        return map(create_foreign_key, self._platform.get_table_foreign_keys(table, database))

    def get_table_foreign_key_names(self, table, database=None, **kwargs):
        return map(ForeignKey.get_name, self.get_table_foreign_keys(table, database, **kwargs))



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

from abc import ABCMeta, abstractmethod

from pydbal.exception import DBALTypesError


class BaseType:
    __metaclass__ = ABCMeta

    ARRAY = "array"
    BOOLEAN = "boolean"
    SMALLINT = "smallint"
    INTEGER = "integer"
    BIGINT = "bigint"
    DECIMAL = "decimal"
    FLOAT = "float"
    STRING = "string"
    TEXT = "text"
    BINARY = "binary"
    BLOB = "blob"
    DATE = "date"
    TIME = "time"
    DATETIME = "datetime"
    GUID = "guid"

    def __init__(self):
        pass

    @staticmethod
    def get_type(name):
        if name not in BaseType.TYPES:
            raise DBALTypesError.unknown_type(name)
        return BaseType.TYPES[name]

    @staticmethod
    @abstractmethod
    def get_name():
        pass


class ArrayType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.ARRAY


class BooleanType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.BOOLEAN


class SmallIntType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.SMALLINT


class IntegerType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.INTEGER


class BigIntType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.BIGINT


class DecimalType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.DECIMAL


class FloatType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.FLOAT


class StringType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.STRING


class TextType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.TEXT


class BinaryType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.BINARY


class BlobType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.BLOB


class DateType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.DATE


class TimeType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.TIME


class DateTimeType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.DATETIME


class GuidType(BaseType):
    @staticmethod
    def get_name():
        return BaseType.GUID


BaseType.TYPES = {
    BaseType.ARRAY: ArrayType,
    BaseType.BOOLEAN: BooleanType,
    BaseType.SMALLINT: SmallIntType,
    BaseType.INTEGER: IntegerType,
    BaseType.BIGINT: BigIntType,
    BaseType.DECIMAL: DecimalType,
    BaseType.FLOAT: FloatType,
    BaseType.STRING: StringType,
    BaseType.TEXT: TextType,
    BaseType.BINARY: BinaryType,
    BaseType.BLOB: BlobType,
    BaseType.DATE: DateType,
    BaseType.TIME: TimeType,
    BaseType.DATETIME: DateTimeType,
    BaseType.GUID: GuidType
}



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement


class DBALError(Exception):
    pass


class DBALConnectionError(DBALError):
    @classmethod
    def unknown_driver(cls, unknown_driver_name, known_driver_names):
        return cls(
            "The given driver '%s' is unknown, pyDBAL currently supports only the following drivers: %s." %
            (unknown_driver_name, ", ".join(sorted(known_driver_names))))

    @classmethod
    def connection_closed(cls):
        return cls("Connection with database is closed.")

    @classmethod
    def no_active_transaction(cls):
        return cls("There is no active transaction.")

    @classmethod
    def commit_failed_rollback_only(cls):
        return cls("Transaction commit failed because the transaction has been marked for rollback only.")

    @classmethod
    def savepoints_not_supported(cls):
        return cls("Savepoints are not supported by this driver.")

    @classmethod
    def may_not_alter_nested_transaction_with_savepoints_in_transaction(cls):
        return cls("May not alter the nested transaction with savepoints behavior while a transaction is open.")


class DBALDriverError(DBALError):
    def __init__(self, *args, **kwargs):
        self._driver_exception = kwargs.pop("exception")
        args = ("%s: %s." % (args[0], self._driver_exception), ) + args[1:]
        super(DBALDriverError, self).__init__(*args, **kwargs)

    def get_driver_exception(self):
        return self._driver_exception

    @classmethod
    def driver_exception(cls, driver, exception):
        return cls("An exception occurred in driver '%s'" % driver.get_name(), exception=exception)

    @classmethod
    def execute_exception(cls, driver, exception, sql, params=None):
        message = "An exception occurred in driver '%s' while executing '%s'" % (driver.get_name(), sql)
        if params:
            message += " with parameters " + str(list(params))
        return cls(message, exception=exception)


class DBALPlatformError(DBALError):
    @classmethod
    def not_supported(cls, method):
        return cls("Operation '%s' is not supported by platform." % method.__name__)

    @classmethod
    def invalid_isolation_level(cls, level):
        return cls("Invalid isolation level '%s'." % level)

    @classmethod
    def invalid_offset(cls, offset):
        return cls("LIMIT argument offset '%s' is not valid." % offset)

    @classmethod
    def offset_not_supported(cls, platform):
        return cls("Platform '%s' does not support offset values in limit queries." % platform.get_name())

    @classmethod
    def unknown_column_type(cls, type_):
        return cls("Unknown database type '%s' requested." % type_)


class DBALStatementError(DBALError):
    @classmethod
    def missing_positional_parameter(cls, param_index, params):
        return cls(
            "Value for positional parameter with index '%d' not found in params array: %s." % (param_index, params))

    @classmethod
    def missing_named_parameter(cls, param_name, params):
        return cls("Value for named parameter ':%s' not found in params array: %s." % (param_name, params))


class DBALBuilderError(DBALError):
    @classmethod
    def unknown_alias(cls, alias, registered_aliases):
        return cls(
            "The given alias '%s' is not part of any FROM or JOIN clause table. "
            "The currently registered aliases are: %s." % (alias, ", ".join(registered_aliases)))

    @classmethod
    def non_unique_alias(cls, alias, registered_aliases):
        return cls(
            "The given alias '%s' is not unique in FROM and JOIN clause table. "
            "The currently registered aliases are: %s." % (alias, ", ".join(registered_aliases)))


class DBALTypesError(DBALError):
    @classmethod
    def unknown_type(cls, name):
        return cls("Unknown column type '%s' requested." % name)


class DBALWarning(Warning):
    pass


class DBALNotImplementedWarning(DBALWarning):
    pass



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import MySQLdb
import MySQLdb.cursors

from pydbal.drivers import BaseDriver
from pydbal.exception import DBALDriverError
from pydbal.platforms.mysql import MySQLPlatform


class MySQLDriver(BaseDriver):
    _cursor = None

    def __init__(self, host, user=None, password=None, database=None, port=3306, timeout=0, charset="utf8",
                 timezone="SYSTEM", sql_mode="TRADITIONAL", **params):
        self._logger = params.pop("logger")
        self._platform = MySQLPlatform(self)

        self._params = dict(
            use_unicode=True, charset=charset, init_command=("SET time_zone = '%s'" % timezone),
            connect_timeout=timeout, sql_mode=sql_mode, autocommit=params.pop("auto_commit"),
            cursorclass=MySQLdb.cursors.SSCursor, **params)

        if user is not None:
            self._params["user"] = user
        if password is not None:
            self._params["passwd"] = password
        if database is not None:
            self._params["db"] = database

        if "/" in host:
            self._params["unix_socket"] = host
        else:
            self._params["host"] = host
            self._params["port"] = int(port)

    def _get_server_version_info(self):
        return getattr(self._conn, "_server_version", None)

    def get_database(self):
        if "db" in self._params:
            return self._params["db"]
        self.execute("SELECT DATABASE()")
        return next(self.iterate())[0][1]

    def connect(self):
        self.close()
        try:
            self._conn = MySQLdb.connect(**self._params)
        except Exception as ex:
            raise DBALDriverError.driver_exception(self, ex)

    def close(self):
        self.clear()
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def clear(self):
        if self._cursor is not None:
            self._cursor.close()
            self._cursor = None

    def error_code(self):
        return self._conn.errno()

    def error_info(self):
        return self._conn.error()

    def execute(self, sql, *params):
        try:
            return self._execute(sql, params)
        except MySQLdb.DatabaseError as ex:
            raise DBALDriverError.execute_exception(self, ex, sql, params)

    def _execute(self, sql, params):
        """Execute statement with reconnecting by connection closed error codes.

        2006 (CR_SERVER_GONE_ERROR): MySQL server has gone away
        2013 (CR_SERVER_LOST): Lost connection to MySQL server during query
        2055 (CR_SERVER_LOST_EXTENDED): Lost connection to MySQL server at '%s', system error: %d
        """
        try:
            return self._execute_unsafe(sql, params)
        except MySQLdb.OperationalError as ex:
            if ex.args[0] in (2006, 2013, 2055):
                self._log("Connection with server is lost. Trying to reconnect.")
                self.connect()
                return self._execute_unsafe(sql, params)
            raise  # MySQLdb.OperationalError

    def _execute_unsafe(self, sql, params):
        self._log(sql, *params)
        self._cursor = self._conn.cursor()
        return self._cursor.execute(sql, params)

    def iterate(self):
        if self._cursor is None:
            raise StopIteration

        columns = [x[0] for x in self._cursor.description]
        for row in self._cursor:
            yield zip(columns, row)

        self.clear()

    def row_count(self):
        return self._conn.affected_rows()

    def last_insert_id(self, seq_name=None):
        return self._conn.insert_id() or None

    def begin_transaction(self):
        self.execute_and_clear("START TRANSACTION")

    def commit(self):
        self._log("COMMIT")
        self._conn.commit()

    def rollback(self):
        self._log("ROLLBACK")
        self._conn.rollback()

    def escape_string(self, value):
        return self._conn.literal(value)

    def get_name(self):
        return "mysql"



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import re

from abc import ABCMeta, abstractmethod

from pydbal.connection import Connection
from pydbal.exception import DBALPlatformError


class BasePlatform:
    __metaclass__ = ABCMeta

    _re_table_column_type = re.compile(r"^(?P<type>\w*)\s*(?:\(\s*(?P<length>\d+(?:,\d+)?)\s*\))?")
    _re_comment_type = re.compile(r"\s*\(DBALType:(?P<type>\w+)\)\s*")
    _keywords = None

    def __init__(self, driver):
        self._driver = driver

    def _fetch(self, sql):
        self._driver.execute(sql)
        return self._driver.iterate()

    def get_keywords(self):
        if self._keywords is None:
            self._keywords = self._get_keywords()
        return self._keywords

    @abstractmethod
    def _get_keywords(self):
        pass

    @abstractmethod
    def _get_type_mappings(self):
        pass

    def get_type_mapping(self, type_):
        mappings = self._get_type_mappings()
        if type_ not in mappings:
            raise DBALPlatformError.unknown_column_type(type_)
        return mappings[type_]

    @staticmethod
    def get_type_from_comment(comment, default=None):
        def replace(match):
            replace._type = match.group("type")
            return ""

        replace._type = default
        comment = BasePlatform._re_comment_type.sub(replace, comment)
        return comment, replace._type

    def quote_identifier(self, identifier):
        return ".".join(map(self.quote_single_identifier, identifier.split(".")))

    def quote_single_identifier(self, identifier):
        c = self.get_identifier_quote_character()
        return c + identifier.replace(c, c + c) + c

    def get_identifier_quote_character(self):
        return '"'

    def modify_limit_sql(self, sql, limit, offset=None):
        if limit is not None:
            limit = int(limit)
        if offset is not None:
            offset = int(offset)
            if offset < 0:
                raise DBALPlatformError.invalid_offset(offset)
            if offset > 0 and not self.is_limit_offset_supported():
                raise DBALPlatformError.offset_not_supported(self)
        return self._modify_limit_sql(sql, limit, offset)

    def _modify_limit_sql(self, sql, limit, offset):
        if limit is not None:
            sql += " LIMIT " + str(limit)
        if offset is not None:
            sql += " OFFSET " + str(offset)
        return sql

    @staticmethod
    def is_limit_offset_supported():
        return True

    @staticmethod
    def is_savepoints_supported():
        return True

    def is_release_savepoints_supported(self):
        return self.is_savepoints_supported()

    @staticmethod
    def is_foreign_keys_supported():
        return True

    def create_savepoint(self, savepoint):
        self._driver.execute_and_clear("SAVEPOINT " + savepoint)

    def release_savepoint(self, savepoint):
        self._driver.execute_and_clear("RELEASE SAVEPOINT " + savepoint)

    def rollback_savepoint(self, savepoint):
        self._driver.execute_and_clear("ROLLBACK TO SAVEPOINT " + savepoint)

    def _get_transaction_isolation_sql(self, level):
        if level == Connection.TRANSACTION_READ_UNCOMMITTED:
            return "READ UNCOMMITTED"
        elif level == Connection.TRANSACTION_READ_COMMITTED:
            return "READ COMMITTED"
        elif level == Connection.TRANSACTION_REPEATABLE_READ:
            return "REPEATABLE READ"
        elif level == Connection.TRANSACTION_SERIALIZABLE:
            return "SERIALIZABLE"
        raise DBALPlatformError.invalid_isolation_level(level)

    def set_transaction_isolation(self, level):
        raise DBALPlatformError.not_supported(self.set_transaction_isolation)

    @staticmethod
    def get_default_transaction_isolation_level():
        return Connection.TRANSACTION_READ_COMMITTED

    def get_databases(self):
        raise DBALPlatformError.not_supported(self.get_databases)

    def get_views(self, database=None):
        raise DBALPlatformError.not_supported(self.get_views)

    def get_tables(self, database=None):
        raise DBALPlatformError.not_supported(self.get_tables)

    def get_table_columns(self, table, database=None):
        raise DBALPlatformError.not_supported(self.get_table_columns)

    def get_table_indexes(self, table, database=None):
        raise DBALPlatformError.not_supported(self.get_table_indexes)

    def get_table_foreign_keys(self, table, database=None):
        raise DBALPlatformError.not_supported(self.get_table_foreign_keys)



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import itertools

from pydbal.platforms import BasePlatform
from pydbal.types import BaseType


class MySQLPlatform(BasePlatform):
    _KEYWORDS = (
        "ADD", "ALL", "ALTER", "ANALYZE", "AND", "AS", "ASC", "ASENSITIVE", "BEFORE", "BETWEEN", "BIGINT", "BINARY",
        "BLOB", "BOTH", "BY", "CALL", "CASCADE", "CASE", "CHANGE", "CHAR", "CHARACTER", "CHECK", "COLLATE", "COLUMN",
        "CONDITION", "CONNECTION", "CONSTRAINT", "CONTINUE", "CONVERT", "CREATE", "CROSS", "CURRENT_DATE",
        "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "CURSOR", "DATABASE", "DATABASES", "DAY_HOUR",
        "DAY_MICROSECOND", "DAY_MINUTE", "DAY_SECOND", "DEC", "DECIMAL", "DECLARE", "DEFAULT", "DELAYED", "DELETE",
        "DESC", "DESCRIBE", "DETERMINISTIC", "DISTINCT", "DISTINCTROW", "DIV", "DOUBLE", "DROP", "DUAL", "EACH", "ELSE",
        "ELSEIF", "ENCLOSED", "ESCAPED", "EXISTS", "EXIT", "EXPLAIN", "FALSE", "FETCH", "FLOAT", "FLOAT4", "FLOAT8",
        "FOR", "FORCE", "FOREIGN", "FROM", "FULLTEXT", "GOTO", "GRANT", "GROUP", "HAVING", "HIGH_PRIORITY",
        "HOUR_MICROSECOND", "HOUR_MINUTE", "HOUR_SECOND", "IF", "IGNORE", "IN", "INDEX", "INFILE", "INNER", "INOUT",
        "INSENSITIVE", "INSERT", "INT", "INT1", "INT2", "INT3", "INT4", "INT8", "INTEGER", "INTERVAL", "INTO", "IS",
        "ITERATE", "JOIN", "KEY", "KEYS", "KILL", "LABEL", "LEADING", "LEAVE", "LEFT", "LIKE", "LIMIT", "LINES", "LOAD",
        "LOCALTIME", "LOCALTIMESTAMP", "LOCK", "LONG", "LONGBLOB", "LONGTEXT", "LOOP", "LOW_PRIORITY", "MATCH",
        "MEDIUMBLOB", "MEDIUMINT", "MEDIUMTEXT", "MIDDLEINT", "MINUTE_MICROSECOND", "MINUTE_SECOND", "MOD", "MODIFIES",
        "NATURAL", "NOT", "NO_WRITE_TO_BINLOG", "NULL", "NUMERIC", "ON", "OPTIMIZE", "OPTION", "OPTIONALLY", "OR",
        "ORDER", "OUT", "OUTER", "OUTFILE", "PRECISION", "PRIMARY", "PROCEDURE", "PURGE", "RAID0", "RANGE", "READ",
        "READS", "REAL", "REFERENCES", "REGEXP", "RELEASE", "RENAME", "REPEAT", "REPLACE", "REQUIRE", "RESTRICT",
        "RETURN", "REVOKE", "RIGHT", "RLIKE", "SCHEMA", "SCHEMAS", "SECOND_MICROSECOND", "SELECT", "SENSITIVE",
        "SEPARATOR", "SET", "SHOW", "SMALLINT", "SONAME", "SPATIAL", "SPECIFIC", "SQL", "SQLEXCEPTION", "SQLSTATE",
        "SQLWARNING", "SQL_BIG_RESULT", "SQL_CALC_FOUND_ROWS", "SQL_SMALL_RESULT", "SSL", "STARTING", "STRAIGHT_JOIN",
        "TABLE", "TERMINATED", "THEN", "TINYBLOB", "TINYINT", "TINYTEXT", "TO", "TRAILING", "TRIGGER", "TRUE", "UNDO",
        "UNION", "UNIQUE", "UNLOCK", "UNSIGNED", "UPDATE", "USAGE", "USE", "USING", "UTC_DATE", "UTC_TIME",
        "UTC_TIMESTAMP", "VALUES", "VARBINARY", "VARCHAR", "VARCHARACTER", "VARYING", "WHEN", "WHERE", "WHILE", "WITH",
        "WRITE", "X509", "XOR", "YEAR_MONTH", "ZEROFILL"
    )
    _KEYWORDS57 = (
        "ACCESSIBLE", "ADD", "ALL", "ALTER", "ANALYZE", "AND", "AS", "ASC", "ASENSITIVE", "BEFORE", "BETWEEN", "BIGINT",
        "BINARY", "BLOB", "BOTH", "BY", "CALL", "CASCADE", "CASE", "CHANGE", "CHAR", "CHARACTER", "CHECK", "COLLATE",
        "COLUMN", "CONDITION", "CONSTRAINT", "CONTINUE", "CONVERT", "CREATE", "CROSS", "CURRENT_DATE", "CURRENT_TIME",
        "CURRENT_TIMESTAMP", "CURRENT_USER", "CURSOR", "DATABASE", "DATABASES", "DAY_HOUR", "DAY_MICROSECOND",
        "DAY_MINUTE", "DAY_SECOND", "DEC", "DECIMAL", "DECLARE", "DEFAULT", "DELAYED", "DELETE", "DESC", "DESCRIBE",
        "DETERMINISTIC", "DISTINCT", "DISTINCTROW", "DIV", "DOUBLE", "DROP", "DUAL", "EACH", "ELSE", "ELSEIF",
        "ENCLOSED", "ESCAPED", "EXISTS", "EXIT", "EXPLAIN", "FALSE", "FETCH", "FLOAT", "FLOAT4", "FLOAT8", "FOR",
        "FORCE", "FOREIGN", "FROM", "FULLTEXT", "GET", "GRANT", "GROUP", "HAVING", "HIGH_PRIORITY", "HOUR_MICROSECOND",
        "HOUR_MINUTE", "HOUR_SECOND", "IF", "IGNORE", "IN", "INDEX", "INFILE", "INNER", "INOUT", "INSENSITIVE",
        "INSERT", "INT", "INT1", "INT2", "INT3", "INT4", "INT8", "INTEGER", "INTERVAL", "INTO", "IO_AFTER_GTIDS",
        "IO_BEFORE_GTIDS", "IS", "ITERATE", "JOIN", "KEY", "KEYS", "KILL", "LEADING", "LEAVE", "LEFT", "LIKE", "LIMIT",
        "LINEAR", "LINES", "LOAD", "LOCALTIME", "LOCALTIMESTAMP", "LOCK", "LONG", "LONGBLOB", "LONGTEXT", "LOOP",
        "LOW_PRIORITY", "MASTER_BIND", "MASTER_SSL_VERIFY_SERVER_CERT", "MATCH", "MAXVALUE", "MEDIUMBLOB", "MEDIUMINT",
        "MEDIUMTEXT", "MIDDLEINT", "MINUTE_MICROSECOND", "MINUTE_SECOND", "MOD", "MODIFIES", "NATURAL",
        "NO_WRITE_TO_BINLOG", "NONBLOCKING", "NOT", "NULL", "NUMERIC", "ON", "OPTIMIZE", "OPTION", "OPTIONALLY", "OR",
        "ORDER", "OUT", "OUTER", "OUTFILE", "PARTITION", "PRECISION", "PRIMARY", "PROCEDURE", "PURGE", "RANGE", "READ",
        "READ_WRITE", "READS", "REAL", "REFERENCES", "REGEXP", "RELEASE", "RENAME", "REPEAT", "REPLACE", "REQUIRE",
        "RESIGNAL", "RESTRICT", "RETURN", "REVOKE", "RIGHT", "RLIKE", "SCHEMA", "SCHEMAS", "SECOND_MICROSECOND",
        "SELECT", "SENSITIVE", "SEPARATOR", "SET", "SHOW", "SIGNAL", "SMALLINT", "SPATIAL", "SPECIFIC", "SQL",
        "SQL_BIG_RESULT", "SQL_CALC_FOUND_ROWS", "SQL_SMALL_RESULT", "SQLEXCEPTION", "SQLSTATE", "SQLWARNING", "SSL",
        "STARTING", "STRAIGHT_JOIN", "TABLE", "TERMINATED", "THEN", "TINYBLOB", "TINYINT", "TINYTEXT", "TO", "TRAILING",
        "TRIGGER", "TRUE", "UNDO", "UNION", "UNIQUE", "UNLOCK", "UNSIGNED", "UPDATE", "USAGE", "USE", "USING",
        "UTC_DATE", "UTC_TIME", "UTC_TIMESTAMP", "VALUES", "VARBINARY", "VARCHAR", "VARCHARACTER", "VARYING", "WHEN",
        "WHERE", "WHILE", "WITH", "WRITE", "XOR", "YEAR_MONTH", "ZEROFILL"
    )

    _TYPE_MAPPINGS = {
        "tinyint": BaseType.BOOLEAN,
        "smallint": BaseType.SMALLINT,
        "mediumint": BaseType.INTEGER,
        "int": BaseType.INTEGER,
        "integer": BaseType.INTEGER,
        "bigint": BaseType.BIGINT,
        "tinytext": BaseType.TEXT,
        "mediumtext": BaseType.TEXT,
        "longtext": BaseType.TEXT,
        "text": BaseType.TEXT,
        "varchar": BaseType.STRING,
        "string": BaseType.STRING,
        "char": BaseType.STRING,
        "date": BaseType.DATE,
        "datetime": BaseType.DATETIME,
        "timestamp": BaseType.DATETIME,
        "time": BaseType.TIME,
        "float": BaseType.FLOAT,
        "double": BaseType.FLOAT,
        "real": BaseType.FLOAT,
        "decimal": BaseType.DECIMAL,
        "numeric": BaseType.DECIMAL,
        "year": BaseType.DATE,
        "longblob": BaseType.BLOB,
        "blob": BaseType.BLOB,
        "mediumblob": BaseType.BLOB,
        "tinyblob": BaseType.BLOB,
        "binary": BaseType.BINARY,
        "varbinary": BaseType.BINARY,
        "set": BaseType.ARRAY
    }

    LENGTH_LIMIT_TINYTEXT = 255
    LENGTH_LIMIT_TEXT = 65535
    LENGTH_LIMIT_MEDIUMTEXT = 16777215
    LENGTH_LIMIT_TINYBLOB = 255
    LENGTH_LIMIT_BLOB = 65535
    LENGTH_LIMIT_MEDIUMBLOB = 16777215

    def _get_keywords(self):
        if self._driver.get_server_version() >= (5, 7):
            return MySQLPlatform._KEYWORDS57
        return MySQLPlatform._KEYWORDS

    def _get_type_mappings(self):
        return MySQLPlatform._TYPE_MAPPINGS

    def get_identifier_quote_character(self):
        return "`"

    def _modify_limit_sql(self, sql, limit, offset):
        if limit is not None:
            sql += " LIMIT " + str(limit)
            if offset is not None:
                sql += " OFFSET " + str(offset)
        elif offset is not None:
            sql += " LIMIT 18446744073709551615 OFFSET " + str(offset)
        return sql

    def set_transaction_isolation(self, level):
        self._driver.execute_and_clear(
            "SET SESSION TRANSACTION ISOLATION LEVEL " + self._get_transaction_isolation_sql(level))

    def get_databases(self):
        for row in self._fetch("SHOW DATABASES"):
            yield row[0][1]  # {"Database": ...}

    def get_views(self, database=None):
        database = "DATABASE()" if database is None else "'" + database + "'"
        sql = "SELECT TABLE_NAME, VIEW_DEFINITION FROM INFORMATION_SCHEMA.VIEWS WHERE TABLE_SCHEMA = " + database + ""
        for row in self._fetch(sql):  # [{"TABLE_NAME": ..., "VIEW_DEFINITION": ...}, ...]
            yield row[0][1], row[1][1]

    def get_tables(self, database=None):
        for row in self._fetch("SHOW FULL TABLES WHERE Table_type = 'BASE TABLE'"):
            yield row[0][1]

    def get_table_columns(self, table, database=None):
        database = "DATABASE()" if database is None else "'" + database + "'"
        sql = "SELECT COLUMN_NAME, COLUMN_TYPE, IS_NULLABLE, COLUMN_DEFAULT, EXTRA, COLUMN_COMMENT, COLLATION_NAME " \
              "FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = " + database + " AND TABLE_NAME = '" + table + "'"

        for row in self._fetch(sql):
            row = dict(row)

            type_match = BasePlatform._re_table_column_type.match(row["COLUMN_TYPE"])
            length = type_match.group("length")

            options = {}

            type_ = type_match.group("type")
            if type_ in ("char", "binary"):
                options["fixed"] = True
            elif type_ in ("float", "double", "real", "numeric", "decimal") and length is not None:
                decimal = (length + ",0").split(",")
                options["precision"] = int(decimal[0])
                options["scale"] = int(decimal[1])
                length = None
            elif type_ == "tinytext":
                length = MySQLPlatform.LENGTH_LIMIT_TINYTEXT
            elif type_ == "text":
                length = MySQLPlatform.LENGTH_LIMIT_TEXT
            elif type_ == "mediumtext":
                length = MySQLPlatform.LENGTH_LIMIT_MEDIUMTEXT
            elif type_ == "tinyblob":
                length = MySQLPlatform.LENGTH_LIMIT_TINYBLOB
            elif type_ == "blob":
                length = MySQLPlatform.LENGTH_LIMIT_BLOB
            elif type_ == "mediumblob":
                length = MySQLPlatform.LENGTH_LIMIT_MEDIUMBLOB
            elif type_ in ("tinyint", "smallint", "mediumint", "int", "integer", "bigint", "year"):
                length = None

            if length:
                options["length"] = int(length)
            if "unsigned" in row["COLUMN_TYPE"]:
                options["unsigned"] = True
            if row["IS_NULLABLE"] == "YES":
                options["notnull"] = False
            if row["COLUMN_DEFAULT"] is not None:
                options["default"] = row["COLUMN_DEFAULT"]
            if "auto_increment" in row["EXTRA"]:
                options["autoincrement"] = True
            if row["COLLATION_NAME"] is not None:
                options["platform_options"] = {"collation": row["COLLATION_NAME"]}

            type_ = self.get_type_mapping(type_)
            if row["COLUMN_COMMENT"]:
                comment, c_type = BasePlatform.get_type_from_comment(row["COLUMN_COMMENT"])
                if comment:
                    options["comment"] = row["COLUMN_COMMENT"]
                if c_type:
                    type_ = c_type

            yield row["COLUMN_NAME"], type_, options

    def get_table_indexes(self, table, database=None):
        database = "DATABASE()" if database is None else "'" + database + "'"
        sql = "SELECT INDEX_NAME, COLUMN_NAME, INDEX_TYPE, NON_UNIQUE FROM INFORMATION_SCHEMA.STATISTICS " \
              "WHERE TABLE_SCHEMA = " + database + " AND TABLE_NAME = '" + table + "'"

        indexes = []
        for row in self._fetch(sql):
            row = dict(row)

            options = {}
            if not row["NON_UNIQUE"]:
                options["unique"] = True
            if row["INDEX_NAME"] == "PRIMARY":
                options["primary"] = True

            if "FULLTEXT" in row["INDEX_TYPE"]:
                options["flags"] = ("FULLTEXT", )
            elif "SPATIAL" in row["INDEX_TYPE"]:
                options["flags"] = ("SPATIAL", )

            indexes.append((row["INDEX_NAME"], row["COLUMN_NAME"], options))

        for group, generator in itertools.groupby(indexes, lambda x: (x[0], x[2])):
            yield group[0], tuple(x[1] for x in generator), group[1]

    def get_table_foreign_keys(self, table, database=None):
        database = "DATABASE()" if database is None else "'" + database + "'"
        sql = "SELECT DISTINCT k.CONSTRAINT_NAME, k.COLUMN_NAME, k.REFERENCED_TABLE_NAME, k.REFERENCED_COLUMN_NAME " \
              "/*!50116 , c.UPDATE_RULE, c.DELETE_RULE */ FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE k " \
              "/*!50116 INNER JOIN INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS c " \
              "ON c.CONSTRAINT_NAME = k.CONSTRAINT_NAME AND c.TABLE_NAME = '" + table + "' */ " \
              "WHERE k.TABLE_NAME = '" + table + "' AND k.TABLE_SCHEMA = " + database + " " \
              "/*!50116 AND c.CONSTRAINT_SCHEMA = " + database + " */ AND k.REFERENCED_COLUMN_NAME IS NOT NULL"

        foreign_keys = []
        for row in self._fetch(sql):
            row = dict(row)

            options = {}
            delete_rule = row.get("DELETE_RULE")
            if delete_rule not in (None, "RESTRICT"):
                options["on_delete"] = delete_rule
            update_rule = row.get("UPDATE_RULE")
            if update_rule not in (None, "RESTRICT"):
                options["on_update"] = update_rule

            foreign_keys.append((
                row["CONSTRAINT_NAME"],
                row["COLUMN_NAME"],
                row["REFERENCED_TABLE_NAME"],
                row["REFERENCED_COLUMN_NAME"],
                options
            ))

        for group, generator in itertools.groupby(foreign_keys, lambda x: (x[0], x[2], x[4])):
            gen1, gen2 = itertools.tee(generator)
            yield group[0], tuple(x[1] for x in gen1), group[1], tuple(x[3] for x in gen2), group[2]



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import time

from threading import RLock
from contextlib import contextmanager
from pydbal.connection import Connection as UnsafeConnection


class SafeConnection:
    POOL_MAX_SIZE = 4
    LOCK = RLock()

    def __init__(self, *args, **kwargs):
        """Initialises thread-safe database connection.

        :param _pool_max_size: connection pool capacity (defaults to POOL_MAX_SIZE)
        """
        self._args = args
        self._kwargs = kwargs

        self._pool = {}
        self._pool_size = 0
        self._pool_max_size = kwargs.pop("_pool_max_size", SafeConnection.POOL_MAX_SIZE)

        assert self._pool_max_size > 0

    def _get_connection(self):
        with SafeConnection.LOCK:
            for conn, locked in self._pool.iteritems():
                if not locked:
                    return conn

            if self._pool_size == self._pool_max_size:
                while True:
                    time.sleep(1)
                    for conn, locked in self._pool.iteritems():
                        if not locked:
                            return conn

            conn = UnsafeConnection(*self._args, **self._kwargs)
            self._pool[conn] = False
            self._pool_size += 1
            return conn

    def _lock(self, conn):
        self._pool[conn] = True

    def _unlock(self, conn):
        self._pool[conn] = False

    @contextmanager
    def locked(self):
        """Context generator for `with` statement, yields thread-safe connection.

        :return: thread-safe connection
        :rtype: pydbal.connection.Connection
        """
        conn = self._get_connection()
        try:
            self._lock(conn)
            yield conn
        finally:
            self._unlock(conn)

    def query(self, sql, *args, **kwargs):
        """Executes an SQL SELECT query and returns rows generator.

        :param sql: query to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: rows generator
        :rtype: generator
        """
        with self.locked() as conn:
            for row in conn.query(sql, *args, **kwargs):
                yield row

    def execute(self, sql, *args, **kwargs):
        """Executes an SQL INSERT/UPDATE/DELETE query with the given parameters and returns the number of affected rows.

        :param sql: statement to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: number of affected rows
        :rtype: int
        """
        with self.locked() as conn:
            return conn.execute(sql, *args, **kwargs)

    def fetch(self, sql, *args, **kwargs):
        """Executes an SQL SELECT query and returns the first row or `None`.

        :param sql: statement to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: the first row or `None`
        """
        with self.locked() as conn:
            return conn.query(sql, *args, **kwargs).fetch()

    def fetch_all(self, sql, *args, **kwargs):
        """Executes an SQL SELECT query and returns all selected rows.

        :param sql: statement to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: all selected rows
        :rtype: list
        """
        with self.locked() as conn:
            return conn.query(sql, *args, **kwargs).fetch_all()

    def fetch_column(self, sql, *args, **kwargs):
        """Executes an SQL SELECT query and returns the first column of the first row or `None`.

        :param sql: statement to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: the first row of the first column or `None`
        """
        with self.locked() as conn:
            return conn.query(sql, *args, **kwargs).fetch_column()

    def transaction(self, callback):
        """Executes a function in a thread-safe transaction.

        The function gets passed locked Connection instance as an (optional) parameter.

        If an exception occurs during execution of the function or transaction commit,
        the transaction is rolled back and the exception re-thrown.

        :param callback: the function to execute in a transaction
        :return: the value returned by the `callback`
        """
        with self.locked() as conn:
            return conn.transaction(callback)

    def insert(self, table, values):
        """Inserts a table row with specified data.

        :param table: the expression of the table to insert data into, quoted or unquoted
        :param values: a dictionary containing column-value pairs
        :return: last inserted ID
        """
        with self.locked() as conn:
            return conn.insert(table, values)

    def update(self, table, values, identifier):
        """Updates a table row with specified data by given identifier.

        :param table: the expression of the table to update quoted or unquoted
        :param values: a dictionary containing column-value pairs
        :param identifier: the update criteria; a dictionary containing column-value pairs
        :return: the number of affected rows
        :rtype: int
        """
        with self.locked() as conn:
            return conn.update(table, values, identifier)

    def delete(self, table, identifier):
        """Deletes a table row by given identifier.

        :param table: the expression of the table to update quoted or unquoted
        :param identifier: the delete criteria; a dictionary containing column-value pairs
        :return: the number of affected rows
        :rtype: int
        """
        with self.locked() as conn:
            return conn.delete(table, identifier)



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

from abc import ABCMeta, abstractmethod


class BaseDriver:
    __metaclass__ = ABCMeta

    _server_version_info = None

    _logger = None
    _platform = None
    _conn = None

    @abstractmethod
    def __init__(self, **params):
        pass

    def __del__(self):
        self.close()

    def _log(self, log, *params):
        logger = self.get_logger()
        if logger is not None:
            if params:
                log += " " + str(list(params))
            logger.debug(log)

    def get_logger(self):
        try:
            # 'getChild' property available in Python 2.7+
            return self._logger.getChild(self.get_name())
        except AttributeError:
            return self._logger

    def get_platform(self):
        return self._platform

    def get_server_version(self):
        info = self.get_server_version_info()
        if info is None:
            return None
        return ".".join(map(str, info))

    def get_server_version_info(self):
        if not self._server_version_info:
            self._server_version_info = self._get_server_version_info()
        return self._server_version_info

    @abstractmethod
    def _get_server_version_info(self):
        pass

    @abstractmethod
    def get_database(self):
        pass

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def clear(self):
        pass

    def is_connected(self):
        return self._conn is not None

    @abstractmethod
    def error_code(self):
        pass

    @abstractmethod
    def error_info(self):
        pass

    @abstractmethod
    def execute(self, sql, *params):
        pass

    def execute_and_clear(self, sql, *params):
        self.execute(sql, *params)
        self.clear()

    @abstractmethod
    def iterate(self):
        pass

    @abstractmethod
    def row_count(self):
        pass

    @abstractmethod
    def last_insert_id(self, seq_name=None):
        pass

    @abstractmethod
    def begin_transaction(self):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def rollback(self):
        pass

    @staticmethod
    def get_placeholder():
        return "%s"

    @abstractmethod
    def escape_string(self, value):
        pass

    @abstractmethod
    def get_name(self):
        pass



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import logging

from pydbal.drivers import BaseDriver
from pydbal.statement import Statement
from pydbal.schema import SchemaManager
from pydbal.exception import DBALConnectionError
from pydbal.builder import SQLBuilder, ExpressionBuilder


class Connection:
    """pyDBAL generic connection class.

    To open new connection import ``Connection`` from ``pydbal.connection``
    package and initialise ``Connection`` class for a required driver with
    desired parameters.
    """
    DRIVERS = {
        "mysql": "pydbal.drivers.mysql.MySQLDriver",
        "sqlite": "pydbal.drivers.sqlite.SQLiteDriver"
    }

    FETCH_DEFAULT = Statement.FETCH_DEFAULT
    FETCH_TUPLE = Statement.FETCH_TUPLE
    FETCH_LIST = Statement.FETCH_LIST
    FETCH_DICT = Statement.FETCH_DICT
    FETCH_OBJECT = Statement.FETCH_OBJECT
    FETCH_COLUMN = Statement.FETCH_COLUMN

    TRANSACTION_READ_UNCOMMITTED = 1
    TRANSACTION_READ_COMMITTED = 2
    TRANSACTION_REPEATABLE_READ = 3
    TRANSACTION_SERIALIZABLE = 4

    _instance_count = 0

    def __init__(self, driver, auto_connect=True, auto_commit=True, fetch_mode=FETCH_DICT, logger=None, **params):
        """Initialises database connection.

        :param driver: database driver
        :param auto_connect: set connection auto (re)connect
        :param auto_commit: set connection auto commit
        :param fetch_mode: set default fetch mode
        :param logger: driver logger
        :param params: database connection parameters
        """
        if not isinstance(logger, logging.Logger):
            logger = Connection.get_default_logger()
        self._logger = logger

        params = dict(auto_commit=auto_commit, logger=logger, **params)
        if isinstance(driver, BaseDriver):
            self._driver = driver
        else:
            if driver not in Connection.DRIVERS:
                raise DBALConnectionError.unknown_driver(driver, Connection.DRIVERS.iterkeys())
            pkg, cls = Connection.DRIVERS[driver].rsplit(".", 1)
            module = __import__(pkg, fromlist=[cls])
            self._driver = getattr(module, cls)(**params)

        self._params = params

        self._platform = self._driver.get_platform()
        self._schema_manager = SchemaManager(self)

        self._expr = ExpressionBuilder(self)
        self._auto_connect = auto_connect
        self._auto_commit = auto_commit
        self._fetch_mode = fetch_mode

        self._transaction_nesting_level = 0
        self._transaction_isolation_level = None
        self._nest_transactions_with_savepoints = False
        self._is_rollback_only = False

        if auto_connect:
            self.connect()

        Connection._instance_count += 1

    def __del__(self):
        """Closes connection on instance destroy."""
        if hasattr(self, "_driver"):
            Connection._instance_count -= 1
            self.close()

    @staticmethod
    def cache_clear():
        """Clears module cache."""
        from pydbal import cache
        cache.clear()
        del cache

    @staticmethod
    def get_default_logger():
        """Returns default driver logger.

        :return: logger instance
        :rtype: logging.Logger
        """
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter(
            "[%(levelname)1.1s %(asctime)s %(name)s] %(message)s",
            "%y%m%d %H:%M:%S"))

        logger_name = "pydbal"
        if Connection._instance_count > 1:
            logger_name += ":" + str(Connection._instance_count)
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        return logger

    def get_logger(self):
        """Returns connection driver logger.

        :return: logger instance
        :rtype: logging.Logger
        """
        return self._logger

    def get_driver(self):
        """Returns the DBAL driver instance.

        :return: driver instance
        :rtype: pydbal.drivers.BaseDriver
        """
        return self._driver

    def get_platform(self):
        """Returns the DBAL platform instance.

        :return: platform instance
        :rtype: pydbal.platforms.BasePlatform
        """
        return self._platform

    def get_platform_version(self):
        """Returns connected platform version.

        :return: platform version
        :rtype: str
        """
        self.ensure_connected()
        return self._driver.get_server_version()

    def get_platform_version_info(self):
        """Returns connected platform version information.

        :return: platform version info
        :rtype: tuple
        """
        self.ensure_connected()
        return self._driver.get_server_version_info()

    def get_database(self):
        """Gets the name of the database this Connection is connected to.

        :return: database name
        :rtype: str
        """
        self.ensure_connected()
        return self._driver.get_database()

    def connect(self):
        """Opens database connection."""
        self._driver.connect()

    def close(self):
        """Closes database connection."""
        self._driver.close()

    def is_connected(self):
        """Checks whether an actual connection to the database is established.

        :return: `True` if connection is open, `False` otherwise
        :rtype: bool
        """
        return self._driver.is_connected()

    def ensure_connected(self):
        """Ensures database connection is still open."""
        if not self.is_connected():
            if not self._auto_connect:
                raise DBALConnectionError.connection_closed()
            self.connect()

    def get_schema_manager(self):
        """Gets the schema manager that can be used to inspect or change the database schema through the connection.

        :return: schema manager
        :rtype: pydbal.schema.SchemaManager
        """
        return self._schema_manager

    def sql_builder(self):
        """Creates the new SQL builder.

        :return: new SQL builder
        :rtype: pydbal.builder.SQLBuilder
        """
        return SQLBuilder(self)

    def get_expression_builder(self):
        """Gets an expression builder.

        :return: expression builder
        :rtype: pydbal.builder.ExpressionBuilder
        """
        return self._expr

    def get_fetch_mode(self):
        """Returns connection fetch mode.

        :return: connection fetch mode
        :rtype: int
        """
        return self._fetch_mode

    def set_fetch_mode(self, fetch_mode):
        """Sets connection fetch mode.

        :param fetch_mode: one of `Connection.FETCH_*` constants
        """
        self._fetch_mode = fetch_mode

    def query(self, sql, *args, **kwargs):
        """Executes an SQL SELECT query, returning a result set as a Statement object.

        :param sql: query to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: result set as a Statement object
        :rtype: pydbal.statement.Statement
        """
        self.ensure_connected()
        stmt = Statement(self)
        stmt.execute(sql, *args, **kwargs)
        return stmt

    def execute(self, sql, *args, **kwargs):
        """Executes an SQL INSERT/UPDATE/DELETE query with the given parameters and returns the number of affected rows.

        :param sql: statement to execute
        :param args: parameters iterable
        :param kwargs: parameters iterable
        :return: number of affected rows
        :rtype: int
        """
        self.ensure_connected()
        return Statement(self).execute(sql, *args, **kwargs)

    def row_count(self):
        """Returns the number of rows affected by the last DELETE, INSERT, or UPDATE statement.

        If the last executed SQL statement was a SELECT statement, some databases may return the number of rows returned
        by that statement. However, this behaviour is not guaranteed for all databases and should not be relied on for
        portable applications.

        :return: number of rows
        :rtype: int
        """
        self.ensure_connected()
        return self._driver.row_count()

    def last_insert_id(self, seq_name=None):
        """Returns the ID of the last inserted row, or the last value from a sequence object,
        depending on the underlying driver.

        Note: This method may not return a meaningful or consistent result across different drivers, because the
        underlying database may not even support the notion of AUTO_INCREMENT/IDENTITY columns or sequences.

        :param seq_name: name of the sequence object from which the ID should be returned
        :return: representation of the last inserted ID
        """
        self.ensure_connected()
        return self._driver.last_insert_id(seq_name)

    def error_code(self):
        """Fetches the SQLSTATE associated with the last database operation.

        :return: the last error code
        :rtype: int
        """
        self.ensure_connected()
        return self._driver.error_code()

    def error_info(self):
        """Fetches extended error information associated with the last database operation.

        :return: the last error information
        """
        self.ensure_connected()
        return self._driver.error_info()

    def begin_transaction(self):
        """Starts a transaction by suspending auto-commit mode."""
        self.ensure_connected()
        self._transaction_nesting_level += 1
        if self._transaction_nesting_level == 1:
            self._driver.begin_transaction()
        elif self._nest_transactions_with_savepoints:
            self.create_savepoint(self._get_nested_transaction_savepoint_name())

    def commit(self):
        """Commits the current transaction."""
        if self._transaction_nesting_level == 0:
            raise DBALConnectionError.no_active_transaction()
        if self._is_rollback_only:
            raise DBALConnectionError.commit_failed_rollback_only()

        self.ensure_connected()
        if self._transaction_nesting_level == 1:
            self._driver.commit()
        elif self._nest_transactions_with_savepoints:
            self.release_savepoint(self._get_nested_transaction_savepoint_name())

        self._transaction_nesting_level -= 1

        if not self._auto_commit and self._transaction_nesting_level == 0:
            self.begin_transaction()

    def commit_all(self):
        """Commits all current nesting transactions."""
        while self._transaction_nesting_level != 0:
            if not self._auto_commit and self._transaction_nesting_level == 1:
                return self.commit()
            self.commit()

    def rollback(self):
        """Cancels any database changes done during the current transaction."""
        if self._transaction_nesting_level == 0:
            raise DBALConnectionError.no_active_transaction()

        self.ensure_connected()
        if self._transaction_nesting_level == 1:
            self._transaction_nesting_level = 0
            self._driver.rollback()
            self._is_rollback_only = False
            if not self._auto_commit:
                self.begin_transaction()
        elif self._nest_transactions_with_savepoints:
            self.rollback_savepoint(self._get_nested_transaction_savepoint_name())
            self._transaction_nesting_level -= 1
        else:
            self._is_rollback_only = True
            self._transaction_nesting_level -= 1

    def transaction(self, callback):
        """Executes a function in a transaction.

        The function gets passed this Connection instance as an (optional) parameter.

        If an exception occurs during execution of the function or transaction commit,
        the transaction is rolled back and the exception re-thrown.

        :param callback: the function to execute in a transaction
        :return: the value returned by the `callback`
        :raise: Exception
        """
        self.begin_transaction()
        try:
            result = callback(self)
            self.commit()
            return result
        except:
            self.rollback()
            raise

    def is_auto_commit(self):
        """Returns the current auto-commit mode for this connection.

        :return: `True` if auto-commit mode is currently enabled for this connection, `False` otherwise
        :rtype: bool
        """
        return self._auto_commit

    def set_auto_commit(self, auto_commit):
        """Sets auto-commit mode for this connection.

        If a connection is in auto-commit mode, then all its SQL statements will be executed and committed as individual
        transactions. Otherwise, its SQL statements are grouped into transactions that are terminated by a call to
        either the method commit or the method rollback. By default, new connections are in auto-commit mode.

        NOTE: If this method is called during a transaction and the auto-commit mode is changed, the transaction is
        committed. If this method is called and the auto-commit mode is not changed, the call is a no-op.

        :param auto_commit: `True` to enable auto-commit mode; `False` to disable it
        """
        auto_commit = bool(auto_commit)
        if auto_commit == self._auto_commit:
            return

        self._auto_commit = auto_commit

        if self.is_connected() and self._transaction_nesting_level != 0:
            self.commit_all()

    def is_transaction_active(self):
        """Checks whether a transaction is currently active.

        :return: `True` if a transaction is currently active, `False` otherwise
        :rtype: bool
        """
        return self._transaction_nesting_level > 0

    def set_rollback_only(self):
        """Marks the current transaction so that the only possible outcome for the transaction to be rolled back."""
        if self._transaction_nesting_level == 0:
            raise DBALConnectionError.no_active_transaction()
        self._is_rollback_only = True

    def is_rollback_only(self):
        """Checks whether the current transaction is marked for rollback only.

        :return: `True` if a transaction is marked for rollback only, `False` otherwise
        :rtype: bool
        """
        if self._transaction_nesting_level == 0:
            raise DBALConnectionError.no_active_transaction()
        return self._is_rollback_only

    def set_transaction_isolation(self, level):
        """Sets the transaction isolation level.

        :param level: the level to set
        """
        self.ensure_connected()
        self._transaction_isolation_level = level
        self._platform.set_transaction_isolation(level)

    def get_transaction_isolation(self):
        """Returns the currently active transaction isolation level.

        :return: the current transaction isolation level
        :rtype: int
        """
        if self._transaction_isolation_level is None:
            self._transaction_isolation_level = self._platform.get_default_transaction_isolation_level()
        return self._transaction_isolation_level

    def set_nest_transactions_with_savepoints(self, nest_transactions_with_savepoints):
        """Sets if nested transactions should use savepoints.

        :param nest_transactions_with_savepoints: `True` or `False`
        """
        if self._transaction_nesting_level > 0:
            raise DBALConnectionError.may_not_alter_nested_transaction_with_savepoints_in_transaction()
        if not self._platform.is_savepoints_supported():
            raise DBALConnectionError.savepoints_not_supported()
        self._nest_transactions_with_savepoints = bool(nest_transactions_with_savepoints)

    def get_nest_transactions_with_savepoints(self):
        """Returns if nested transactions should use savepoints.

        :return: `True` if should use savepoints, `False` otherwise
        :rtype: bool
        """
        return self._nest_transactions_with_savepoints

    def _get_nested_transaction_savepoint_name(self):
        """Returns the savepoint name to use for nested transactions

        :return: a string with the savepoint name or false
        :rtype: str
        """
        return "PYDBAL_SAVEPOINT_" + str(self._transaction_nesting_level)

    def create_savepoint(self, savepoint):
        """Creates a new savepoint.

        :param savepoint: the name of the savepoint to create
        :raise: pydbal.exception.DBALConnectionError
        """
        if not self._platform.is_savepoints_supported():
            raise DBALConnectionError.savepoints_not_supported()
        self.ensure_connected()
        self._platform.create_savepoint(savepoint)

    def release_savepoint(self, savepoint):
        """Releases the given savepoint.

        :param savepoint: the name of the savepoint to release
        :raise: pydbal.exception.DBALConnectionError
        """
        if not self._platform.is_savepoints_supported():
            raise DBALConnectionError.savepoints_not_supported()
        if self._platform.is_release_savepoints_supported():
            self.ensure_connected()
            self._platform.release_savepoint(savepoint)

    def rollback_savepoint(self, savepoint):
        """Rolls back to the given savepoint.

        :param savepoint: the name of the savepoint to rollback to
        :raise: pydbal.exception.DBALConnectionError
        """
        if not self._platform.is_savepoints_supported():
            raise DBALConnectionError.savepoints_not_supported()
        self.ensure_connected()
        self._platform.rollback_savepoint(savepoint)

    def insert(self, table, values):
        """Inserts a table row with specified data.

        :param table: the expression of the table to insert data into, quoted or unquoted
        :param values: a dictionary containing column-value pairs
        :return: last inserted ID
        """
        assert isinstance(values, dict)

        sb = self.sql_builder().insert(table)
        for column, value in values.iteritems():
            sb.set_value(column, sb.create_positional_parameter(value))
        return sb.execute()

    def update(self, table, values, identifier):
        """Updates a table row with specified data by given identifier.

        :param table: the expression of the table to update quoted or unquoted
        :param values: a dictionary containing column-value pairs
        :param identifier: the update criteria; a dictionary containing column-value pairs
        :return: the number of affected rows
        :rtype: int
        """
        assert isinstance(values, dict)
        assert isinstance(identifier, dict)

        sb = self.sql_builder().update(table)
        for column, value in values.iteritems():
            sb.set(column, sb.create_positional_parameter(value))
        for column, value in identifier.iteritems():
            func = self._expr.in_ if isinstance(value, (list, tuple)) else self._expr.eq
            sb.and_where(func(column, sb.create_positional_parameter(value)))
        return sb.execute()

    def delete(self, table, identifier):
        """Deletes a table row by given identifier.

        :param table: the expression of the table to update quoted or unquoted
        :param identifier: the delete criteria; a dictionary containing column-value pairs
        :return: the number of affected rows
        :rtype: int
        """
        assert isinstance(identifier, dict)

        sb = self.sql_builder().delete(table)
        for column, value in identifier.iteritems():
            func = self._expr.in_ if isinstance(value, (list, tuple)) else self._expr.eq
            sb.and_where(func(column, sb.create_positional_parameter(value)))
        return sb.execute()



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import sqlite3
import warnings

from pydbal.drivers import BaseDriver
from pydbal.exception import DBALDriverError, DBALNotImplementedWarning
from pydbal.platforms.sqlite import SQLitePlatform


class SQLiteDriver(BaseDriver):
    _cursor = None
    _error = None

    def __init__(self, database, timeout=5.0, **params):
        self._logger = params.pop("logger")
        self._platform = SQLitePlatform(self)

        auto_commit = params.pop("auto_commit")
        if auto_commit:
            params["isolation_level"] = None
        else:
            params["isolation_level"] = "EXCLUSIVE"

        self._params = dict(database=database, timeout=timeout, **params)

    def _get_server_version_info(self):
        return sqlite3.sqlite_version_info

    def get_database(self):
        return self._params["database"]

    @staticmethod
    def _row_factory(cursor, row):
        return tuple((cursor.description[i][0], value) for i, value in enumerate(row))

    def connect(self):
        self.close()
        try:
            self._conn = sqlite3.connect(**self._params)
            self._conn.row_factory = self._row_factory
        except Exception as ex:
            raise DBALDriverError.driver_exception(self, ex)

    def close(self):
        self.clear()
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def clear(self):
        if self._cursor is not None:
            self._cursor.close()
            self._cursor = None

    def error_code(self):
        warnings.warn("Python 'sqlite3' library does not expose error codes.", DBALNotImplementedWarning)
        return 1 if self._error else 0

    def error_info(self):
        return self._error

    def execute(self, sql, *params):
        try:
            self._log(sql, *params)
            self._error = None
            self._cursor = self._conn.execute(sql, params)
            return self.row_count()
        except Exception as ex:
            if isinstance(ex, sqlite3.OperationalError):
                self._error = ex.message
            raise DBALDriverError.execute_exception(self, ex, sql, params)

    def iterate(self):
        if self._cursor is None:
            raise StopIteration

        for row in self._cursor:
            yield row

        self.clear()

    def row_count(self):
        return getattr(self._cursor, "rowcount", 0)

    def last_insert_id(self, seq_name=None):
        return getattr(self._cursor, "lastrowid", None)

    def begin_transaction(self):
        self.execute_and_clear("BEGIN TRANSACTION")

    def commit(self):
        self._log("COMMIT")
        self._conn.commit()

    def rollback(self):
        self._log("ROLLBACK")
        self._conn.rollback()

    @staticmethod
    def get_placeholder():
        return "?"

    def escape_string(self, value):
        return "'" + value.replace("'", "''") + "'"

    def get_name(self):
        return "sqlite"



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

try:
    import setuptools
    from setuptools import setup
except ImportError:
    setuptools = None
    from distutils.core import setup

kwargs = {}

version = "0.11"

with open("README.rst") as f:
    kwargs["long_description"] = f.read()

if setuptools is not None:
    install_requires = ["MySQL-python"]
    kwargs["install_requires"] = install_requires

setup(
    name="pydbal",
    version=version,
    packages=["pydbal", "pydbal.drivers", "pydbal.platforms"],
    description="Database Abstraction Layer (DBAL) for Python 2.6+",
    author="Alexander Lokhman",
    author_email="alex.lokhman@gmail.com",
    url="https://github.com/lokhman/pydbal",
    download_url="https://github.com/lokhman/pydbal/tarball/" + version,
    keywords=["database", "abstraction", "sql", "mysql", "sqlite"],
    license="https://opensource.org/licenses/MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ],
    **kwargs
)



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import re
import itertools

from pydbal.platforms import BasePlatform
from pydbal.cache import cached
from pydbal.types import BaseType
from pydbal.connection import Connection
from pydbal.exception import DBALPlatformError


class SQLitePlatform(BasePlatform):
    _KEYWORDS = (
        "ABORT", "ACTION", "ADD", "AFTER", "ALL", "ALTER", "ANALYZE", "AND", "AS", "ASC", "ATTACH", "AUTOINCREMENT",
        "BEFORE", "BEGIN", "BETWEEN", "BY", "CASCADE", "CASE", "CAST", "CHECK", "COLLATE", "COLUMN", "COMMIT",
        "CONFLICT", "CONSTRAINT", "CREATE", "CROSS", "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "DATABASE",
        "DEFAULT", "DEFERRABLE", "DEFERRED", "DELETE", "DESC", "DETACH", "DISTINCT", "DROP", "EACH", "ELSE", "END",
        "ESCAPE", "EXCEPT", "EXCLUSIVE", "EXISTS", "EXPLAIN", "FAIL", "FOR", "FOREIGN", "FROM", "FULL", "GLOB", "GROUP",
        "HAVING", "IF", "IGNORE", "IMMEDIATE", "IN", "INDEX", "INDEXED", "INITIALLY", "INNER", "INSERT", "INSTEAD",
        "INTERSECT", "INTO", "IS", "ISNULL", "JOIN", "KEY", "LEFT", "LIKE", "LIMIT", "MATCH", "NATURAL", "NO", "NOT",
        "NOTNULL", "NULL", "OF", "OFFSET", "ON", "OR", "ORDER", "OUTER", "PLAN", "PRAGMA", "PRIMARY", "QUERY", "RAISE",
        "REFERENCES", "REGEXP", "REINDEX", "RELEASE", "RENAME", "REPLACE", "RESTRICT", "RIGHT", "ROLLBACK", "ROW",
        "SAVEPOINT", "SELECT", "SET", "TABLE", "TEMP", "TEMPORARY", "THEN", "TO", "TRANSACTION", "TRIGGER", "UNION",
        "UNIQUE", "UPDATE", "USING", "VACUUM", "VALUES", "VIEW", "VIRTUAL", "WHEN", "WHERE"
    )

    _TYPE_MAPPINGS = {
        "boolean": BaseType.BOOLEAN,
        "tinyint": BaseType.BOOLEAN,
        "smallint": BaseType.SMALLINT,
        "mediumint": BaseType.INTEGER,
        "int": BaseType.INTEGER,
        "integer": BaseType.INTEGER,
        "serial": BaseType.INTEGER,
        "bigint": BaseType.BIGINT,
        "bigserial": BaseType.BIGINT,
        "clob": BaseType.TEXT,
        "tinytext": BaseType.TEXT,
        "mediumtext": BaseType.TEXT,
        "longtext": BaseType.TEXT,
        "text": BaseType.TEXT,
        "varchar": BaseType.STRING,
        "longvarchar": BaseType.STRING,
        "varchar2": BaseType.STRING,
        "nvarchar": BaseType.STRING,
        "image": BaseType.STRING,
        "ntext": BaseType.STRING,
        "char": BaseType.STRING,
        "date": BaseType.DATE,
        "datetime": BaseType.DATETIME,
        "timestamp": BaseType.DATETIME,
        "time": BaseType.TIME,
        "float": BaseType.FLOAT,
        "double": BaseType.FLOAT,
        "double precision": BaseType.FLOAT,
        "real": BaseType.FLOAT,
        "decimal": BaseType.DECIMAL,
        "numeric": BaseType.DECIMAL,
        "binary": BaseType.BINARY,
        "blob": BaseType.BLOB
    }

    _re_foreign_key_details = re.compile(
        r"(?:CONSTRAINT\s+([^\s]+)\s+)?(?:FOREIGN\s+KEY[^\)]+\)\s*)?REFERENCES\s+[^\s]+\s+(?:\([^\)]+\))?"
        r"(?:[^,]*?(NOT\s+DEFERRABLE|DEFERRABLE)(?:\s+INITIALLY\s+(DEFERRED|IMMEDIATE))?)?", re.IGNORECASE)

    def _get_keywords(self):
        return SQLitePlatform._KEYWORDS

    def _get_type_mappings(self):
        return SQLitePlatform._TYPE_MAPPINGS

    def _modify_limit_sql(self, sql, limit, offset):
        if limit is None and offset is not None:
            return sql + " LIMIT -1 OFFSET " + str(offset)
        return super(SQLitePlatform, self)._modify_limit_sql(sql, limit, offset)

    def _get_transaction_isolation_sql(self, level):
        if level == Connection.TRANSACTION_READ_UNCOMMITTED:
            return 0
        elif level in (
            Connection.TRANSACTION_READ_COMMITTED,
            Connection.TRANSACTION_REPEATABLE_READ,
            Connection.TRANSACTION_SERIALIZABLE
        ):
            return 1
        raise DBALPlatformError.invalid_isolation_level(level)

    def set_transaction_isolation(self, level):
        self._driver.execute_and_clear("PRAGMA read_uncommitted = " + self._get_transaction_isolation_sql(level))

    @staticmethod
    def _escape(name):
        return name.replace(".", "__")

    @cached
    def _fetch_table_create_sql(self, table):
        sql = "SELECT sql FROM (SELECT * FROM sqlite_master UNION ALL SELECT * FROM sqlite_temp_master) " \
              "WHERE type = 'table' AND name = '" + SQLitePlatform._escape(table) + "'"
        return dict(next(self._fetch(sql))).get("sql", "")

    def get_views(self, database=None):
        for row in self._fetch("SELECT name, sql FROM sqlite_master WHERE type = 'view' AND sql NOT NULL"):
            yield row[0][1], row[1][1]  # {"name": ..., "sql": ...}

    def get_tables(self, database=None):
        sql = "SELECT name FROM sqlite_master WHERE type = 'table' AND name != 'sqlite_sequence' " \
              "AND name != 'geometry_columns' AND name != 'spatial_ref_sys' UNION ALL " \
              "SELECT name FROM sqlite_temp_master WHERE type = 'table' ORDER BY name"
        for row in self._fetch(sql):  # [{"name": ...}, ...]
            yield row[0][1]

    def get_table_columns(self, table, database=None):
        create_sql = self._fetch_table_create_sql(table)
        for row in self._fetch("PRAGMA TABLE_INFO('" + SQLitePlatform._escape(table) + "')"):
            row = dict(row)

            type_match = BasePlatform._re_table_column_type.match(row["type"])
            length = type_match.group("length")

            options = {}

            type_ = type_match.group("type").lower()
            if type_ == "char":
                options["fixed"] = True
            elif type_ in ("float", "double", "real", "decimal", "numeric") and length is not None:
                decimal = (length + ",0").split(",")
                options["precision"] = int(decimal[0])
                options["scale"] = int(decimal[1])
                length = None
            elif type_ == "integer" and row["pk"]:
                options["autoincrement"] = True

            if length:
                options["length"] = int(length)
            if "unsigned" in row["type"]:
                options["unsigned"] = True
            if not row["notnull"]:
                options["notnull"] = False
            if row["dflt_value"] not in (None, "NULL"):
                options["default"] = row["dflt_value"]

            name = row["name"].replace("'", "''")
            type_ = self.get_type_mapping(type_)
            if type_ in (BaseType.STRING, BaseType.TEXT):
                re_ = r"(?:" + name + r")[^,(]+(?:\([^()]+\)[^,]*)?(?:(?:DEFAULT|CHECK)\s*(?:\(.*?\))?[^,]*)*" \
                      r"COLLATE\s+[\"']?([^\s,\"')]+)"
                matches = re.findall(re_, create_sql, re.IGNORECASE | re.DOTALL)
                options["platform_options"] = {"collation": matches[0] if matches else "BINARY"}

            re_ = r"[\s(,](?:" + name + r")(?:\(.*?\)|[^,(])*?,?((?:\s*--[^\n]*\n?)+)"
            matches = re.findall(re_, create_sql, re.IGNORECASE | re.DOTALL)
            if matches:
                comment = re.sub(r"^\s*--\s*", "", matches[0].rstrip("\r\n"), flags=re.MULTILINE)
                comment, c_type = BasePlatform.get_type_from_comment(comment)
                if comment:
                    options["comment"] = comment
                if c_type:
                    type_ = c_type

            yield row["name"], type_, options

    def get_table_indexes(self, table, database=None):
        table = SQLitePlatform._escape(table)

        primary = []
        for row in self._fetch("PRAGMA TABLE_INFO('" + table + "')"):
            row = dict(row)
            if row.get("pk"):
                primary.append(row["name"])

        if primary:
            yield "PRIMARY", primary, {
                "primary": True,
                "unique": True
            }

        for row in self._fetch("PRAGMA INDEX_LIST('" + table + "')"):
            row = dict(row)

            name = row["name"]
            if name.startswith("sqlite_"):
                continue

            options = {}
            if row["unique"]:
                options["unique"] = True

            columns = self._fetch("PRAGMA INDEX_INFO('" + SQLitePlatform._escape(name) + "')")
            yield name, [dict(column)["name"] for column in columns], options

    def get_table_foreign_keys(self, table, database=None):
        foreign_keys = []
        for row in self._fetch("PRAGMA FOREIGN_KEY_LIST('" + SQLitePlatform._escape(table) + "')"):
            row = dict(row)

            options = {}
            delete_rule = row.get("on_delete")
            if delete_rule not in (None, "RESTRICT"):
                options["on_delete"] = delete_rule
            update_rule = row.get("on_update")
            if update_rule not in (None, "RESTRICT"):
                options["on_update"] = update_rule

            foreign_keys.append([
                row["id"],
                row["from"],
                row["table"],
                row["to"],
                options
            ])

        if foreign_keys:
            create_sql = self._fetch_table_create_sql(table)
            matches = SQLitePlatform._re_foreign_key_details.findall(create_sql)
            for i, details in enumerate(matches):
                for foreign_key in foreign_keys:
                    if foreign_key[0] != i:
                        continue
                    if details[1].upper() == "DEFERRABLE":
                        foreign_key[4]["deferrable"] = True
                    if details[2].upper() == "DEFERRED":
                        foreign_key[4]["deferred"] = True
                    foreign_key[0] = details[0]

        for group, generator in itertools.groupby(foreign_keys, lambda x: (x[0], x[2], x[4])):
            gen1, gen2 = itertools.tee(generator)
            yield group[0], tuple(x[1] for x in gen1), group[1], tuple(x[3] for x in gen2), group[2]



#!/usr/bin/env python
#
# Copyright (c) 2016 Alexander Lokhman <alex.lokhman@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import, division, print_function, with_statement

import re

from collections import namedtuple

from pydbal.exception import DBALStatementError


class Statement:
    OBJECT_NAME = "Object"

    FETCH_DEFAULT = 0
    FETCH_TUPLE = 1
    FETCH_LIST = 2
    FETCH_DICT = 3
    FETCH_OBJECT = 4
    FETCH_COLUMN = 5

    _re_params = re.compile(
        r"(\?|(?<!:):[a-zA-Z_][a-zA-Z0-9_]*)(?=(?:(?:\\.|[^'\"\\])*['\"](?:\\.|[^'\"\\])*['\"])*(?:\\.|[^'\"\\])*\Z)")

    def __init__(self, connection):
        self._connection = connection

    def __iter__(self):
        return self.iterate()

    def clear(self):
        self._connection.get_driver().clear()

    def iterate(self, fetch_mode=None, column_index=0):
        if fetch_mode is None:
            fetch_mode = self._connection.get_fetch_mode()

        for row in self._connection.get_driver().iterate():
            yield self._transform(row, fetch_mode, column_index)

    @staticmethod
    def _transform(row, fetch_mode=FETCH_DEFAULT, column_index=0):
        if fetch_mode == Statement.FETCH_DICT:
            return dict(row)

        if fetch_mode == Statement.FETCH_COLUMN:
            return row[column_index][1]

        values = tuple(x[1] for x in row)
        if fetch_mode == Statement.FETCH_TUPLE:
            return values
        elif fetch_mode == Statement.FETCH_LIST:
            return list(values)
        elif fetch_mode == Statement.FETCH_OBJECT:
            return namedtuple(Statement.OBJECT_NAME, [x[0] for x in row])._make(values)

        return row

    def execute(self, sql, *args, **kwargs):
        for i, arg in enumerate(args):
            kwargs[i] = arg
        params = []
        sql = Statement._re_params.sub(self._prepare(kwargs, params), sql)
        return self._connection.get_driver().execute(sql, *params)

    def _prepare(self, params, exec_params):
        def replace(match):
            key = match.group()
            if key == "?":
                key = replace._param_counter
                replace._param_counter += 1
            else:
                key = key.lstrip(":")

            if key not in params:
                if isinstance(key, int):
                    raise DBALStatementError.missing_positional_parameter(key, params)
                else:
                    raise DBALStatementError.missing_named_parameter(key, params)

            param = params[key]

            if isinstance(param, (list, tuple)):
                exec_params.extend(param)
                return ", ".join((replace._placeholder, ) * len(param))

            exec_params.append(param)
            return replace._placeholder

        replace._placeholder = self._connection.get_driver().get_placeholder()
        replace._param_counter = 0

        return replace

    def fetch(self, fetch_mode=None):
        try:
            return next(self.iterate(fetch_mode))
        except StopIteration:
            return None

    def fetch_all(self, fetch_mode=None, column_index=0):
        return list(self.iterate(fetch_mode, column_index))

    def fetch_column(self, column_index=0):
        try:
            return next(self.iterate(Statement.FETCH_COLUMN, column_index))
        except StopIteration:
            return None
